# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/exchange-info/api.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import empty_pb2 as google_dot_protobuf_dot_empty__pb2
from galts_trade_api.transport.protos.exchange_info import entities_pb2 as protos_dot_exchange__info_dot_entities__pb2
from galts_trade_api.transport.protos import common_pb2 as protos_dot_common__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/exchange-info/api.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n\x1eprotos/exchange-info/api.proto\x1a\x1bgoogle/protobuf/empty.proto\x1a#protos/exchange-info/entities.proto\x1a\x13protos/common.proto\"\x10\n\x02Id\x12\n\n\x02id\x18\x01 \x01(\x05\"0\n\x0b\x45ntityKinds\x12!\n\x0c\x65ntity_kinds\x18\x01 \x03(\x0e\x32\x0b.EntityKind\"\x96\x03\n\x08\x45ntities\x12%\n\x06\x61ssets\x18\x01 \x03(\x0b\x32\x15.Entities.AssetsEntry\x12\'\n\x07symbols\x18\x02 \x03(\x0b\x32\x16.Entities.SymbolsEntry\x12\'\n\x07markets\x18\x03 \x03(\x0b\x32\x16.Entities.MarketsEntry\x12+\n\texchanges\x18\x04 \x03(\x0b\x32\x18.Entities.ExchangesEntry\x1a\x35\n\x0b\x41ssetsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x15\n\x05value\x18\x02 \x01(\x0b\x32\x06.Asset:\x02\x38\x01\x1a\x37\n\x0cSymbolsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x16\n\x05value\x18\x02 \x01(\x0b\x32\x07.Symbol:\x02\x38\x01\x1a\x37\n\x0cMarketsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x16\n\x05value\x18\x02 \x01(\x0b\x32\x07.Market:\x02\x38\x01\x1a;\n\x0e\x45xchangesEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\x18\n\x05value\x18\x02 \x01(\x0b\x32\t.Exchange:\x02\x38\x01\":\n\nAddAssetIn\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03tag\x18\x02 \x01(\t\x12\x11\n\tprecision\x18\x03 \x01(\x05\"<\n\x0b\x41\x64\x64SymbolIn\x12\x15\n\rbase_asset_id\x18\x01 \x01(\x05\x12\x16\n\x0equote_asset_id\x18\x02 \x01(\x05\"a\n\x0b\x41\x64\x64MarketIn\x12\x12\n\ncustom_tag\x18\x01 \x01(\t\x12\x11\n\tsymbol_id\x18\x02 \x01(\x05\x12\x13\n\x0b\x65xchange_id\x18\x03 \x01(\x05\x12\x16\n\x0etrade_endpoint\x18\x04 \x01(\t\"*\n\rAddExchangeIn\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0b\n\x03tag\x18\x02 \x01(\t\"V\n\x1bSubscribeToEntitiesChangeIn\x12\x14\n\x0csession_hash\x18\x01 \x01(\t\x12!\n\x0c\x65ntity_kinds\x18\x02 \x03(\x0e\x32\x0b.EntityKind\"\xbb\x01\n\x1cSubscribeToEntitiesChangeOut\x12\x0f\n\x07\x63ounter\x18\x01 \x01(\x03\x12\x14\n\x0c\x63ounter_prev\x18\x02 \x01(\x03\x12\x17\n\x05\x61sset\x18\x03 \x01(\x0b\x32\x06.AssetH\x00\x12\x19\n\x06symbol\x18\x04 \x01(\x0b\x32\x07.SymbolH\x00\x12\x19\n\x06market\x18\x05 \x01(\x0b\x32\x07.MarketH\x00\x12\x1d\n\x08\x65xchange\x18\x06 \x01(\x0b\x32\t.ExchangeH\x00\x42\x06\n\x04kind*=\n\nEntityKind\x12\t\n\x05\x41SSET\x10\x00\x12\n\n\x06SYMBOL\x10\x01\x12\n\n\x06MARKET\x10\x02\x12\x0c\n\x08\x45XCHANGE\x10\x03\x32\xfb\x05\n\x0f\x45xchangeInfoApi\x12\x34\n\x0cStartSession\x12\x16.google.protobuf.Empty\x1a\x0c.SessionHash\x12*\n\rFinishSession\x12\x0c.SessionHash\x1a\x0b.BoolResult\x12Z\n\x19SubscribeToEntitiesChange\x12\x1c.SubscribeToEntitiesChangeIn\x1a\x1d.SubscribeToEntitiesChangeOut0\x01\x12&\n\x0bGetEntities\x12\x0c.EntityKinds\x1a\t.Entities\x12\x1c\n\x08\x41\x64\x64\x41sset\x12\x0b.AddAssetIn\x1a\x03.Id\x12\x17\n\x08GetAsset\x12\x03.Id\x1a\x06.Asset\x12*\n\x0b\x44\x65leteAsset\x12\x03.Id\x1a\x16.google.protobuf.Empty\x12\x1e\n\tAddSymbol\x12\x0c.AddSymbolIn\x1a\x03.Id\x12\x19\n\tGetSymbol\x12\x03.Id\x1a\x07.Symbol\x12+\n\x0c\x44\x65leteSymbol\x12\x03.Id\x1a\x16.google.protobuf.Empty\x12\x1e\n\tAddMarket\x12\x0c.AddMarketIn\x1a\x03.Id\x12\x19\n\tGetMarket\x12\x03.Id\x1a\x07.Market\x12+\n\x0c\x44\x65leteMarket\x12\x03.Id\x1a\x16.google.protobuf.Empty\x12\"\n\x0b\x41\x64\x64\x45xchange\x12\x0e.AddExchangeIn\x1a\x03.Id\x12\x1d\n\x0bGetExchange\x12\x03.Id\x1a\t.Exchange\x12-\n\x0e\x44\x65leteExchange\x12\x03.Id\x1a\x16.google.protobuf.Empty\x12.\n\x0f\x44isableExchange\x12\x03.Id\x1a\x16.google.protobuf.Empty\x12-\n\x0e\x45nableExchange\x12\x03.Id\x1a\x16.google.protobuf.Emptyb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_empty__pb2.DESCRIPTOR,protos_dot_exchange__info_dot_entities__pb2.DESCRIPTOR,protos_dot_common__pb2.DESCRIPTOR,])

_ENTITYKIND = _descriptor.EnumDescriptor(
  name='EntityKind',
  full_name='EntityKind',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ASSET', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SYMBOL', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='MARKET', index=2, number=2,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='EXCHANGE', index=3, number=3,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=1141,
  serialized_end=1202,
)
_sym_db.RegisterEnumDescriptor(_ENTITYKIND)

EntityKind = enum_type_wrapper.EnumTypeWrapper(_ENTITYKIND)
ASSET = 0
SYMBOL = 1
MARKET = 2
EXCHANGE = 3



_ID = _descriptor.Descriptor(
  name='Id',
  full_name='Id',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Id.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=121,
  serialized_end=137,
)


_ENTITYKINDS = _descriptor.Descriptor(
  name='EntityKinds',
  full_name='EntityKinds',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='entity_kinds', full_name='EntityKinds.entity_kinds', index=0,
      number=1, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=139,
  serialized_end=187,
)


_ENTITIES_ASSETSENTRY = _descriptor.Descriptor(
  name='AssetsEntry',
  full_name='Entities.AssetsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Entities.AssetsEntry.key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Entities.AssetsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=368,
  serialized_end=421,
)

_ENTITIES_SYMBOLSENTRY = _descriptor.Descriptor(
  name='SymbolsEntry',
  full_name='Entities.SymbolsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Entities.SymbolsEntry.key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Entities.SymbolsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=423,
  serialized_end=478,
)

_ENTITIES_MARKETSENTRY = _descriptor.Descriptor(
  name='MarketsEntry',
  full_name='Entities.MarketsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Entities.MarketsEntry.key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Entities.MarketsEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=480,
  serialized_end=535,
)

_ENTITIES_EXCHANGESENTRY = _descriptor.Descriptor(
  name='ExchangesEntry',
  full_name='Entities.ExchangesEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='Entities.ExchangesEntry.key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='value', full_name='Entities.ExchangesEntry.value', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=_b('8\001'),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=537,
  serialized_end=596,
)

_ENTITIES = _descriptor.Descriptor(
  name='Entities',
  full_name='Entities',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='assets', full_name='Entities.assets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbols', full_name='Entities.symbols', index=1,
      number=2, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='markets', full_name='Entities.markets', index=2,
      number=3, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exchanges', full_name='Entities.exchanges', index=3,
      number=4, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[_ENTITIES_ASSETSENTRY, _ENTITIES_SYMBOLSENTRY, _ENTITIES_MARKETSENTRY, _ENTITIES_EXCHANGESENTRY, ],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=190,
  serialized_end=596,
)


_ADDASSETIN = _descriptor.Descriptor(
  name='AddAssetIn',
  full_name='AddAssetIn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='AddAssetIn.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tag', full_name='AddAssetIn.tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='precision', full_name='AddAssetIn.precision', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=598,
  serialized_end=656,
)


_ADDSYMBOLIN = _descriptor.Descriptor(
  name='AddSymbolIn',
  full_name='AddSymbolIn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='base_asset_id', full_name='AddSymbolIn.base_asset_id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quote_asset_id', full_name='AddSymbolIn.quote_asset_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=658,
  serialized_end=718,
)


_ADDMARKETIN = _descriptor.Descriptor(
  name='AddMarketIn',
  full_name='AddMarketIn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='custom_tag', full_name='AddMarketIn.custom_tag', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol_id', full_name='AddMarketIn.symbol_id', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exchange_id', full_name='AddMarketIn.exchange_id', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trade_endpoint', full_name='AddMarketIn.trade_endpoint', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=720,
  serialized_end=817,
)


_ADDEXCHANGEIN = _descriptor.Descriptor(
  name='AddExchangeIn',
  full_name='AddExchangeIn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='AddExchangeIn.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tag', full_name='AddExchangeIn.tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=819,
  serialized_end=861,
)


_SUBSCRIBETOENTITIESCHANGEIN = _descriptor.Descriptor(
  name='SubscribeToEntitiesChangeIn',
  full_name='SubscribeToEntitiesChangeIn',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='session_hash', full_name='SubscribeToEntitiesChangeIn.session_hash', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='entity_kinds', full_name='SubscribeToEntitiesChangeIn.entity_kinds', index=1,
      number=2, type=14, cpp_type=8, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=863,
  serialized_end=949,
)


_SUBSCRIBETOENTITIESCHANGEOUT = _descriptor.Descriptor(
  name='SubscribeToEntitiesChangeOut',
  full_name='SubscribeToEntitiesChangeOut',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='counter', full_name='SubscribeToEntitiesChangeOut.counter', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='counter_prev', full_name='SubscribeToEntitiesChangeOut.counter_prev', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='asset', full_name='SubscribeToEntitiesChangeOut.asset', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='SubscribeToEntitiesChangeOut.symbol', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='market', full_name='SubscribeToEntitiesChangeOut.market', index=4,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exchange', full_name='SubscribeToEntitiesChangeOut.exchange', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='kind', full_name='SubscribeToEntitiesChangeOut.kind',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=952,
  serialized_end=1139,
)

_ENTITYKINDS.fields_by_name['entity_kinds'].enum_type = _ENTITYKIND
_ENTITIES_ASSETSENTRY.fields_by_name['value'].message_type = protos_dot_exchange__info_dot_entities__pb2._ASSET
_ENTITIES_ASSETSENTRY.containing_type = _ENTITIES
_ENTITIES_SYMBOLSENTRY.fields_by_name['value'].message_type = protos_dot_exchange__info_dot_entities__pb2._SYMBOL
_ENTITIES_SYMBOLSENTRY.containing_type = _ENTITIES
_ENTITIES_MARKETSENTRY.fields_by_name['value'].message_type = protos_dot_exchange__info_dot_entities__pb2._MARKET
_ENTITIES_MARKETSENTRY.containing_type = _ENTITIES
_ENTITIES_EXCHANGESENTRY.fields_by_name['value'].message_type = protos_dot_exchange__info_dot_entities__pb2._EXCHANGE
_ENTITIES_EXCHANGESENTRY.containing_type = _ENTITIES
_ENTITIES.fields_by_name['assets'].message_type = _ENTITIES_ASSETSENTRY
_ENTITIES.fields_by_name['symbols'].message_type = _ENTITIES_SYMBOLSENTRY
_ENTITIES.fields_by_name['markets'].message_type = _ENTITIES_MARKETSENTRY
_ENTITIES.fields_by_name['exchanges'].message_type = _ENTITIES_EXCHANGESENTRY
_SUBSCRIBETOENTITIESCHANGEIN.fields_by_name['entity_kinds'].enum_type = _ENTITYKIND
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['asset'].message_type = protos_dot_exchange__info_dot_entities__pb2._ASSET
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['symbol'].message_type = protos_dot_exchange__info_dot_entities__pb2._SYMBOL
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['market'].message_type = protos_dot_exchange__info_dot_entities__pb2._MARKET
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['exchange'].message_type = protos_dot_exchange__info_dot_entities__pb2._EXCHANGE
_SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind'].fields.append(
  _SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['asset'])
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['asset'].containing_oneof = _SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind']
_SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind'].fields.append(
  _SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['symbol'])
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['symbol'].containing_oneof = _SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind']
_SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind'].fields.append(
  _SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['market'])
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['market'].containing_oneof = _SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind']
_SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind'].fields.append(
  _SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['exchange'])
_SUBSCRIBETOENTITIESCHANGEOUT.fields_by_name['exchange'].containing_oneof = _SUBSCRIBETOENTITIESCHANGEOUT.oneofs_by_name['kind']
DESCRIPTOR.message_types_by_name['Id'] = _ID
DESCRIPTOR.message_types_by_name['EntityKinds'] = _ENTITYKINDS
DESCRIPTOR.message_types_by_name['Entities'] = _ENTITIES
DESCRIPTOR.message_types_by_name['AddAssetIn'] = _ADDASSETIN
DESCRIPTOR.message_types_by_name['AddSymbolIn'] = _ADDSYMBOLIN
DESCRIPTOR.message_types_by_name['AddMarketIn'] = _ADDMARKETIN
DESCRIPTOR.message_types_by_name['AddExchangeIn'] = _ADDEXCHANGEIN
DESCRIPTOR.message_types_by_name['SubscribeToEntitiesChangeIn'] = _SUBSCRIBETOENTITIESCHANGEIN
DESCRIPTOR.message_types_by_name['SubscribeToEntitiesChangeOut'] = _SUBSCRIBETOENTITIESCHANGEOUT
DESCRIPTOR.enum_types_by_name['EntityKind'] = _ENTITYKIND
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Id = _reflection.GeneratedProtocolMessageType('Id', (_message.Message,), {
  'DESCRIPTOR' : _ID,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:Id)
  })
_sym_db.RegisterMessage(Id)

EntityKinds = _reflection.GeneratedProtocolMessageType('EntityKinds', (_message.Message,), {
  'DESCRIPTOR' : _ENTITYKINDS,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:EntityKinds)
  })
_sym_db.RegisterMessage(EntityKinds)

Entities = _reflection.GeneratedProtocolMessageType('Entities', (_message.Message,), {

  'AssetsEntry' : _reflection.GeneratedProtocolMessageType('AssetsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENTITIES_ASSETSENTRY,
    '__module__' : 'protos.exchange_info.api_pb2'
    # @@protoc_insertion_point(class_scope:Entities.AssetsEntry)
    })
  ,

  'SymbolsEntry' : _reflection.GeneratedProtocolMessageType('SymbolsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENTITIES_SYMBOLSENTRY,
    '__module__' : 'protos.exchange_info.api_pb2'
    # @@protoc_insertion_point(class_scope:Entities.SymbolsEntry)
    })
  ,

  'MarketsEntry' : _reflection.GeneratedProtocolMessageType('MarketsEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENTITIES_MARKETSENTRY,
    '__module__' : 'protos.exchange_info.api_pb2'
    # @@protoc_insertion_point(class_scope:Entities.MarketsEntry)
    })
  ,

  'ExchangesEntry' : _reflection.GeneratedProtocolMessageType('ExchangesEntry', (_message.Message,), {
    'DESCRIPTOR' : _ENTITIES_EXCHANGESENTRY,
    '__module__' : 'protos.exchange_info.api_pb2'
    # @@protoc_insertion_point(class_scope:Entities.ExchangesEntry)
    })
  ,
  'DESCRIPTOR' : _ENTITIES,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:Entities)
  })
_sym_db.RegisterMessage(Entities)
_sym_db.RegisterMessage(Entities.AssetsEntry)
_sym_db.RegisterMessage(Entities.SymbolsEntry)
_sym_db.RegisterMessage(Entities.MarketsEntry)
_sym_db.RegisterMessage(Entities.ExchangesEntry)

AddAssetIn = _reflection.GeneratedProtocolMessageType('AddAssetIn', (_message.Message,), {
  'DESCRIPTOR' : _ADDASSETIN,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:AddAssetIn)
  })
_sym_db.RegisterMessage(AddAssetIn)

AddSymbolIn = _reflection.GeneratedProtocolMessageType('AddSymbolIn', (_message.Message,), {
  'DESCRIPTOR' : _ADDSYMBOLIN,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:AddSymbolIn)
  })
_sym_db.RegisterMessage(AddSymbolIn)

AddMarketIn = _reflection.GeneratedProtocolMessageType('AddMarketIn', (_message.Message,), {
  'DESCRIPTOR' : _ADDMARKETIN,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:AddMarketIn)
  })
_sym_db.RegisterMessage(AddMarketIn)

AddExchangeIn = _reflection.GeneratedProtocolMessageType('AddExchangeIn', (_message.Message,), {
  'DESCRIPTOR' : _ADDEXCHANGEIN,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:AddExchangeIn)
  })
_sym_db.RegisterMessage(AddExchangeIn)

SubscribeToEntitiesChangeIn = _reflection.GeneratedProtocolMessageType('SubscribeToEntitiesChangeIn', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIBETOENTITIESCHANGEIN,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:SubscribeToEntitiesChangeIn)
  })
_sym_db.RegisterMessage(SubscribeToEntitiesChangeIn)

SubscribeToEntitiesChangeOut = _reflection.GeneratedProtocolMessageType('SubscribeToEntitiesChangeOut', (_message.Message,), {
  'DESCRIPTOR' : _SUBSCRIBETOENTITIESCHANGEOUT,
  '__module__' : 'protos.exchange_info.api_pb2'
  # @@protoc_insertion_point(class_scope:SubscribeToEntitiesChangeOut)
  })
_sym_db.RegisterMessage(SubscribeToEntitiesChangeOut)


_ENTITIES_ASSETSENTRY._options = None
_ENTITIES_SYMBOLSENTRY._options = None
_ENTITIES_MARKETSENTRY._options = None
_ENTITIES_EXCHANGESENTRY._options = None

_EXCHANGEINFOAPI = _descriptor.ServiceDescriptor(
  name='ExchangeInfoApi',
  full_name='ExchangeInfoApi',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  serialized_start=1205,
  serialized_end=1968,
  methods=[
  _descriptor.MethodDescriptor(
    name='StartSession',
    full_name='ExchangeInfoApi.StartSession',
    index=0,
    containing_service=None,
    input_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    output_type=protos_dot_common__pb2._SESSIONHASH,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='FinishSession',
    full_name='ExchangeInfoApi.FinishSession',
    index=1,
    containing_service=None,
    input_type=protos_dot_common__pb2._SESSIONHASH,
    output_type=protos_dot_common__pb2._BOOLRESULT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='SubscribeToEntitiesChange',
    full_name='ExchangeInfoApi.SubscribeToEntitiesChange',
    index=2,
    containing_service=None,
    input_type=_SUBSCRIBETOENTITIESCHANGEIN,
    output_type=_SUBSCRIBETOENTITIESCHANGEOUT,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetEntities',
    full_name='ExchangeInfoApi.GetEntities',
    index=3,
    containing_service=None,
    input_type=_ENTITYKINDS,
    output_type=_ENTITIES,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='AddAsset',
    full_name='ExchangeInfoApi.AddAsset',
    index=4,
    containing_service=None,
    input_type=_ADDASSETIN,
    output_type=_ID,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetAsset',
    full_name='ExchangeInfoApi.GetAsset',
    index=5,
    containing_service=None,
    input_type=_ID,
    output_type=protos_dot_exchange__info_dot_entities__pb2._ASSET,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteAsset',
    full_name='ExchangeInfoApi.DeleteAsset',
    index=6,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='AddSymbol',
    full_name='ExchangeInfoApi.AddSymbol',
    index=7,
    containing_service=None,
    input_type=_ADDSYMBOLIN,
    output_type=_ID,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetSymbol',
    full_name='ExchangeInfoApi.GetSymbol',
    index=8,
    containing_service=None,
    input_type=_ID,
    output_type=protos_dot_exchange__info_dot_entities__pb2._SYMBOL,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteSymbol',
    full_name='ExchangeInfoApi.DeleteSymbol',
    index=9,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='AddMarket',
    full_name='ExchangeInfoApi.AddMarket',
    index=10,
    containing_service=None,
    input_type=_ADDMARKETIN,
    output_type=_ID,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetMarket',
    full_name='ExchangeInfoApi.GetMarket',
    index=11,
    containing_service=None,
    input_type=_ID,
    output_type=protos_dot_exchange__info_dot_entities__pb2._MARKET,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteMarket',
    full_name='ExchangeInfoApi.DeleteMarket',
    index=12,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='AddExchange',
    full_name='ExchangeInfoApi.AddExchange',
    index=13,
    containing_service=None,
    input_type=_ADDEXCHANGEIN,
    output_type=_ID,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='GetExchange',
    full_name='ExchangeInfoApi.GetExchange',
    index=14,
    containing_service=None,
    input_type=_ID,
    output_type=protos_dot_exchange__info_dot_entities__pb2._EXCHANGE,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DeleteExchange',
    full_name='ExchangeInfoApi.DeleteExchange',
    index=15,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='DisableExchange',
    full_name='ExchangeInfoApi.DisableExchange',
    index=16,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
  _descriptor.MethodDescriptor(
    name='EnableExchange',
    full_name='ExchangeInfoApi.EnableExchange',
    index=17,
    containing_service=None,
    input_type=_ID,
    output_type=google_dot_protobuf_dot_empty__pb2._EMPTY,
    serialized_options=None,
  ),
])
_sym_db.RegisterServiceDescriptor(_EXCHANGEINFOAPI)

DESCRIPTOR.services_by_name['ExchangeInfoApi'] = _EXCHANGEINFOAPI

# @@protoc_insertion_point(module_scope)



import pickle
from typing import Dict, Sequence, Set, Tuple


class Singleton(type):
    """
    Requirement to return an old object are determined by calculating hash of a constructor call
    arguments. One issue with this logic is that if the constructor in second time was called
    with keyword-style of some arguments, the logic will fail if it was called with only
    position-style of the same arguments. In this situation a new object will be created.
    """

    _instances = {}
    _hash_args = []
    _hash_kwargs = {}

    def __new__(mcs, *args, singleton_hash_args=None, singleton_hash_kwargs=None, **kwargs):
        mcs._hash_args = singleton_hash_args if singleton_hash_args is not None else []
        mcs._hash_kwargs = singleton_hash_kwargs if singleton_hash_kwargs is not None else {}

        return super().__new__(mcs, *args, **kwargs)

    def __call__(cls, *args, **kwargs):
        args_hash = cls._calculate_args_hash(args, kwargs)
        key = (cls, args_hash)

        if key not in cls._instances:
            cls._instances[key] = super().__call__(*args, **kwargs)

        return cls._instances[key]

    @classmethod
    def _calculate_args_hash(mcs, args: Tuple, kwargs: Dict) -> int:
        filtered_args = [args[i] for i in range(len(args)) if i in mcs._hash_args]
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in mcs._hash_kwargs}
        hash_source = (filtered_args, filtered_kwargs)

        result = hash(pickle.dumps(hash_source, protocol=pickle.HIGHEST_PROTOCOL))

        return result


def find_duplicates_in_list(source_list: Sequence) -> Set:
    return set([tag for tag in source_list if source_list.count(tag) > 1])



import datetime
from asyncio import Event, wait_for
from decimal import Decimal
from typing import Awaitable, Callable, Dict, List, Mapping, MutableMapping, Optional, Tuple, Union

from .asset import Asset, Symbol
from .exchange import Exchange, Market
from .tools import find_duplicates_in_list
from .transport import DepthConsumeKey, TransportFactory

PriceLevelWithoutFee = Tuple[Decimal, Decimal]
PriceLevelWithFee = Tuple[Decimal, Decimal, Optional[Decimal]]
PriceLevel = Union[PriceLevelWithoutFee, PriceLevelWithFee]
PriceDepth = List[PriceLevel]

OnPriceCallable = Callable[[str, str, str, datetime.datetime, PriceDepth, PriceDepth], Awaitable]


class Terminal:
    def __init__(self, transport: TransportFactory):
        self._transport_factory: TransportFactory = transport
        self._exchange_entities_inited = Event()
        self._assets_by_id: Dict[int, Asset] = {}
        self._assets_by_tag: Dict[str, Asset] = {}
        self._symbols_by_id: Dict[int, Symbol] = {}
        self._symbols_by_tag: Dict[str, Symbol] = {}
        self._exchanges_by_id: Dict[int, Exchange] = {}
        self._exchanges_by_tag: Dict[str, Exchange] = {}

    @property
    def transport_factory(self):
        return self._transport_factory

    @transport_factory.setter
    def transport_factory(self, value):
        self._transport_factory = value

    @property
    def assets_by_id(self):
        return self._assets_by_id

    @property
    def assets_by_tag(self):
        return self._assets_by_tag

    @property
    def symbols_by_id(self):
        return self._symbols_by_id

    @property
    def symbols_by_tag(self):
        return self._symbols_by_tag

    @property
    def exchanges_by_id(self):
        return self._exchanges_by_id

    @property
    def exchanges_by_tag(self):
        return self._exchanges_by_tag

    async def init_transport(self, loop_debug: Optional[bool] = None) -> None:
        await self.transport_factory.init(loop_debug)

    def shutdown_transport(self) -> None:
        self.transport_factory.shutdown()

    async def auth_user(self, username: str, password: str) -> bool:
        return True

    def is_exchange_entities_inited(self) -> bool:
        return self._exchange_entities_inited.is_set()

    async def wait_exchange_entities_inited(self, timeout: float = 5.0) -> None:
        await wait_for(self._exchange_entities_inited.wait(), timeout)

    async def init_exchange_entities(self) -> None:
        await self.transport_factory.get_exchange_entities(
            self._on_init_exchange_entities_response
        )

    async def subscribe_to_prices(
        self,
        callback: OnPriceCallable,
        consume_keys: Optional[List[DepthConsumeKey]] = None
    ) -> None:
        await self.transport_factory.consume_price_depth(
            lambda event: callback(*event),
            consume_keys
        )

    async def _on_init_exchange_entities_response(self, data: MutableMapping[str, Mapping]) -> None:
        properties_to_fill = ('exchanges', 'markets', 'symbols', 'assets')

        for prop in properties_to_fill:
            if prop not in data:
                raise KeyError(f'Key "{prop}" is required')

            data[prop] = {k: v for k, v in data[prop].items() if not v['delete_time']}

        all_assets_tags = [entity['tag'] for entity in data['assets'].values()]
        duplicates = find_duplicates_in_list(all_assets_tags)
        if len(duplicates):
            raise ValueError(f"Assets with duplicates in tags found: {', '.join(duplicates)}")

        for entity in data['assets'].values():
            asset = Asset(**entity)
            self._assets_by_id[entity['id']] = asset
            self._assets_by_tag[entity['tag']] = asset

        for entity in data['symbols'].values():
            if entity['base_asset_id'] not in data['assets']:
                raise ValueError(
                    f"No base asset with id {entity['base_asset_id']} "
                    f"has been found for symbol with id {entity['id']}"
                )
            if entity['quote_asset_id'] not in data['assets']:
                raise ValueError(
                    f"No quote asset with id {entity['quote_asset_id']} "
                    f"has been found for symbol with id {entity['id']}"
                )

            symbol = Symbol(**entity)
            base_asset = self._assets_by_id[symbol.base_asset_id]
            quote_asset = self._assets_by_id[symbol.quote_asset_id]
            tag = Symbol.form_tag(base_asset.tag, quote_asset.tag)

            if tag in self._symbols_by_tag:
                raise ValueError(f'Symbols with duplicates in tags found: {tag}')

            self._symbols_by_id[entity['id']] = symbol
            self._symbols_by_tag[tag] = symbol

        all_exchanges_tags = [entity['tag'] for entity in data['exchanges'].values()]
        duplicates = find_duplicates_in_list(all_exchanges_tags)
        if len(duplicates):
            raise ValueError(f"Exchanges with duplicates in tags found: {', '.join(duplicates)}")

        for entity in data['exchanges'].values():
            exchange = Exchange(**entity)
            self._exchanges_by_id[entity['id']] = exchange
            self._exchanges_by_tag[entity['tag']] = exchange

        for entity in data['markets'].values():
            if entity['exchange_id'] not in self._exchanges_by_id:
                raise ValueError(
                    f"No exchange with id {entity['exchange_id']} "
                    f"has been found for market with id {entity['id']}"
                )
            if entity['symbol_id'] not in data['symbols']:
                raise ValueError(
                    f"No symbol with id {entity['symbol_id']} "
                    f"has been found for market with id {entity['id']}"
                )

            self._exchanges_by_id[entity['exchange_id']].add_market(Market(**entity))

        self._exchange_entities_inited.set()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/exchange-info/entities.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='protos/exchange-info/entities.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  serialized_pb=_b('\n#protos/exchange-info/entities.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\xa3\x01\n\x05\x41sset\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0b\n\x03tag\x18\x02 \x01(\t\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x64\x65lete_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x11\n\tprecision\x18\x06 \x01(\x05\" \n\x06\x41ssets\x12\x16\n\x06\x61ssets\x18\x01 \x03(\x0b\x32\x06.Asset\"\xa5\x01\n\x06Symbol\x12\n\n\x02id\x18\x01 \x01(\x05\x12/\n\x0b\x63reate_time\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x64\x65lete_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x15\n\rbase_asset_id\x18\x04 \x01(\x05\x12\x16\n\x0equote_asset_id\x18\x05 \x01(\x05\"\xca\x01\n\x06Market\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x12\n\ncustom_tag\x18\x02 \x01(\t\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x64\x65lete_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x11\n\tsymbol_id\x18\x05 \x01(\x05\x12\x13\n\x0b\x65xchange_id\x18\x06 \x01(\x05\x12\x16\n\x0etrade_endpoint\x18\x07 \x01(\t\"\xc5\x01\n\x08\x45xchange\x12\n\n\x02id\x18\x01 \x01(\x05\x12\x0b\n\x03tag\x18\x02 \x01(\t\x12/\n\x0b\x63reate_time\x18\x03 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12/\n\x0b\x64\x65lete_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04name\x18\x05 \x01(\t\x12\x30\n\x0c\x64isable_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.Timestampb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,])




_ASSET = _descriptor.Descriptor(
  name='Asset',
  full_name='Asset',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Asset.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tag', full_name='Asset.tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Asset.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete_time', full_name='Asset.delete_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='Asset.name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='precision', full_name='Asset.precision', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=73,
  serialized_end=236,
)


_ASSETS = _descriptor.Descriptor(
  name='Assets',
  full_name='Assets',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='assets', full_name='Assets.assets', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=238,
  serialized_end=270,
)


_SYMBOL = _descriptor.Descriptor(
  name='Symbol',
  full_name='Symbol',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Symbol.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Symbol.create_time', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete_time', full_name='Symbol.delete_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='base_asset_id', full_name='Symbol.base_asset_id', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='quote_asset_id', full_name='Symbol.quote_asset_id', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=273,
  serialized_end=438,
)


_MARKET = _descriptor.Descriptor(
  name='Market',
  full_name='Market',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Market.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='custom_tag', full_name='Market.custom_tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Market.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete_time', full_name='Market.delete_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='symbol_id', full_name='Market.symbol_id', index=4,
      number=5, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='exchange_id', full_name='Market.exchange_id', index=5,
      number=6, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='trade_endpoint', full_name='Market.trade_endpoint', index=6,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=441,
  serialized_end=643,
)


_EXCHANGE = _descriptor.Descriptor(
  name='Exchange',
  full_name='Exchange',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Exchange.id', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='tag', full_name='Exchange.tag', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='create_time', full_name='Exchange.create_time', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='delete_time', full_name='Exchange.delete_time', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='name', full_name='Exchange.name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='disable_time', full_name='Exchange.disable_time', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=646,
  serialized_end=843,
)

_ASSET.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ASSET.fields_by_name['delete_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ASSETS.fields_by_name['assets'].message_type = _ASSET
_SYMBOL.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_SYMBOL.fields_by_name['delete_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MARKET.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_MARKET.fields_by_name['delete_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXCHANGE.fields_by_name['create_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXCHANGE.fields_by_name['delete_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_EXCHANGE.fields_by_name['disable_time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
DESCRIPTOR.message_types_by_name['Asset'] = _ASSET
DESCRIPTOR.message_types_by_name['Assets'] = _ASSETS
DESCRIPTOR.message_types_by_name['Symbol'] = _SYMBOL
DESCRIPTOR.message_types_by_name['Market'] = _MARKET
DESCRIPTOR.message_types_by_name['Exchange'] = _EXCHANGE
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Asset = _reflection.GeneratedProtocolMessageType('Asset', (_message.Message,), {
  'DESCRIPTOR' : _ASSET,
  '__module__' : 'protos.exchange_info.entities_pb2'
  # @@protoc_insertion_point(class_scope:Asset)
  })
_sym_db.RegisterMessage(Asset)

Assets = _reflection.GeneratedProtocolMessageType('Assets', (_message.Message,), {
  'DESCRIPTOR' : _ASSETS,
  '__module__' : 'protos.exchange_info.entities_pb2'
  # @@protoc_insertion_point(class_scope:Assets)
  })
_sym_db.RegisterMessage(Assets)

Symbol = _reflection.GeneratedProtocolMessageType('Symbol', (_message.Message,), {
  'DESCRIPTOR' : _SYMBOL,
  '__module__' : 'protos.exchange_info.entities_pb2'
  # @@protoc_insertion_point(class_scope:Symbol)
  })
_sym_db.RegisterMessage(Symbol)

Market = _reflection.GeneratedProtocolMessageType('Market', (_message.Message,), {
  'DESCRIPTOR' : _MARKET,
  '__module__' : 'protos.exchange_info.entities_pb2'
  # @@protoc_insertion_point(class_scope:Market)
  })
_sym_db.RegisterMessage(Market)

Exchange = _reflection.GeneratedProtocolMessageType('Exchange', (_message.Message,), {
  'DESCRIPTOR' : _EXCHANGE,
  '__module__' : 'protos.exchange_info.entities_pb2'
  # @@protoc_insertion_point(class_scope:Exchange)
  })
_sym_db.RegisterMessage(Exchange)


# @@protoc_insertion_point(module_scope)



import asyncio
from abc import ABC, abstractmethod
from collections.abc import Sequence
from multiprocessing.connection import Connection
from typing import Any, Awaitable, Callable, List, MutableMapping, NamedTuple, Optional, Set

OnExceptionCallable = Callable[[BaseException], Any]


class DepthConsumeKey(NamedTuple):
    exchange_tag: str = '*'
    market_tag: str = '*'
    symbol_tag: str = '*'

    def format_for_rabbitmq(self) -> str:
        required_fields = ('exchange_tag', 'market_tag', 'symbol_tag')

        for arg in required_fields:
            if not len(getattr(self, arg).strip()):
                raise ValueError(f'Field {arg} should be non-empty string')

        if self.exchange_tag == self.market_tag == self.symbol_tag == '*':
            return '#'

        return f'{self.exchange_tag}.{self.market_tag}.{self.symbol_tag}'


class MessageConsumerCollection:
    def __init__(self):
        self._consumers: Set[Callable[..., Awaitable]] = set()

    def add_consumer(self, callback: Callable[..., Awaitable]) -> None:
        self._consumers.add(callback)

    async def send(self, data: Any) -> None:
        coroutines = [consumer(data) for consumer in self._consumers]

        tasks_result = await asyncio.gather(*coroutines, return_exceptions=True)

        for result in tasks_result:
            if isinstance(result, BaseException):
                raise result


class PipeRequest:
    """
    Each subclass of this should be decorated by @dataclass(frozen=True) otherwise
    PipeResponseRouter will fail to distinct request's objects from each other and to select
    an appropriate MessageConsumerCollection object.
    """


class PipeResponseRouter:
    def __init__(self, connection: Connection, poll_delay: float = 0.001):
        self._connection = connection
        self._poll_delay = float(poll_delay)
        self._consumers: MutableMapping[PipeRequest, MessageConsumerCollection] = {}

    async def start(self) -> None:
        while True:
            if not self._connection.poll():
                await asyncio.sleep(self._poll_delay)
                continue

            message = self._connection.recv()

            if not isinstance(message, Sequence):
                raise ValueError('Pipe response message should be an object with indexing')

            if len(message) == 1 and isinstance(message[0], Exception):
                raise message[0]

            if len(message) != 2:
                raise ValueError('Pipe response message should contain exactly 2 elements')

            request, response = message
            self._dispatch(request, response)

    def prepare_consumers_of_response(self, request: PipeRequest) -> MessageConsumerCollection:
        if request not in self._consumers:
            self._consumers[request] = MessageConsumerCollection()

        return self._consumers[request]

    def _dispatch(self, request: PipeRequest, response: Any) -> None:
        if request not in self._consumers:
            return

        # Don't store the sub-tasks because we don't want to cancel them if self.start() has been
        # cancelled.
        asyncio.create_task(self._consumers[request].send(response))


class TransportFactory(ABC):
    async def init(self, loop_debug: Optional[bool] = None) -> None:
        pass

    def shutdown(self) -> None:
        pass

    @abstractmethod
    async def get_exchange_entities(
        self,
        on_response: Callable[..., Awaitable]
    ) -> MessageConsumerCollection:
        pass

    @abstractmethod
    async def consume_price_depth(
        self,
        on_response: Callable[..., Awaitable],
        consume_keys: Optional[List[DepthConsumeKey]] = None
    ) -> MessageConsumerCollection:
        pass


class TransportFactoryException(Exception): pass



import asyncio
import datetime
import json
from dataclasses import dataclass
from decimal import Decimal
from functools import partial
from multiprocessing import Event, Pipe, Process
from multiprocessing.connection import Connection
from typing import Any, Awaitable, Callable, Dict, List, MutableMapping, Optional, Type

import aio_pika

from . import DepthConsumeKey, MessageConsumerCollection, PipeRequest, PipeResponseRouter, \
    TransportFactory, TransportFactoryException
from .exchange_info_client import ExchangeInfoClient
from .grpc_utils import generate_request_id
from ..asyncio_helper import AsyncProgramEnv, run_program_forever
from ..tools import Singleton

AioPikaConsumeCallable = Callable[[aio_pika.IncomingMessage], Any]


@dataclass(frozen=True)
class GetExchangeEntitiesRequest(PipeRequest):
    dsn: str
    timeout: float


@dataclass(frozen=True)
class ConsumePriceDepthRequest(PipeRequest):
    dsn: str
    exchange: str
    consume_keys: frozenset


class RabbitConnection(
    metaclass=Singleton,
    singleton_hash_args=[0],
    singleton_hash_kwargs=['dsn']
):
    def __init__(self, dsn: str):
        self._dsn: str = str(dsn).strip()
        self._connection: Optional[aio_pika.Connection] = None

    @property
    def connection(self):
        return self._connection

    async def create_channel(self, prefetch_count: int = 100) -> aio_pika.Channel:
        if self._connection is None:
            self._connection = await aio_pika.connect_robust(self._dsn)

        channel = await self._connection.channel()
        await channel.set_qos(prefetch_count=prefetch_count)

        return channel


class RabbitConsumer:
    def __init__(
        self,
        channel: aio_pika.Channel,
        exchange_name: str,
        on_message: AioPikaConsumeCallable
    ):
        self._channel = channel
        self._exchange_name = str(exchange_name).strip()
        self._on_message = on_message
        self._exchange: Optional[aio_pika.Exchange] = None
        self._queue: Optional[aio_pika.Queue] = None

    @property
    def channel(self):
        return self._channel

    @property
    def exchange_name(self):
        return self._exchange_name

    @property
    def exchange(self):
        return self._exchange

    @property
    def queue(self):
        return self._queue

    async def create_queue(self) -> aio_pika.Queue:
        self._exchange = await self.channel.declare_exchange(self._exchange_name, passive=True)
        self._queue = await self.channel.declare_queue(exclusive=True)

        await self.queue.consume(self._on_message, no_ack=True)

        return self.queue


class RealTransportFactory(TransportFactory):
    def __init__(
        self,
        exchange_info_dsn: str,
        depth_scraping_queue_dsn: str,
        depth_scraping_queue_exchange: str,
        exchange_info_get_entities_timeout: float = 5.0,
        process_ready_timeout: float = 2.0,
    ):
        super().__init__()
        self._process: Optional[Process] = None
        self._process_ready = Event()
        self._parent_connection, self._child_connection = Pipe()
        self._response_router: Optional[PipeResponseRouter] = None

        def sanity_string(value: str) -> str: return str(value).strip()

        self._exchange_info_dsn = sanity_string(exchange_info_dsn)
        self._exchange_info_get_entities_timeout = float(exchange_info_get_entities_timeout)
        self._depth_scraping_queue_dsn = sanity_string(depth_scraping_queue_dsn)
        self._depth_scraping_queue_exchange = sanity_string(depth_scraping_queue_exchange)
        self._process_ready_timeout = float(process_ready_timeout)

    @property
    def exchange_info_dsn(self):
        return self._exchange_info_dsn

    @property
    def exchange_info_get_entities_timeout(self):
        return self._exchange_info_get_entities_timeout

    @property
    def depth_scraping_queue_dsn(self):
        return self._depth_scraping_queue_dsn

    @property
    def depth_scraping_queue_exchange(self):
        return self._depth_scraping_queue_exchange

    @property
    def process_ready_timeout(self):
        return self._process_ready_timeout

    async def init(self, loop_debug: Optional[bool] = None) -> None:
        if self._process:
            raise RuntimeError('A process for RealTransportFactory should be created only once')

        self._process = RealTransportProcess(
            loop_debug=loop_debug,
            ready_event=self._process_ready,
            connection=self._child_connection
        )
        self._process.start()

        if not self._process_ready.wait(self.process_ready_timeout):
            raise RuntimeError('Failed to initialize RealTransportFactory in time')

        self._response_router = PipeResponseRouter(self._parent_connection)
        task = asyncio.create_task(self._response_router.start())

        def task_done_cb(t: asyncio.Task) -> None:
            if t.cancelled():
                self.shutdown()
                return

            if t.exception():
                self.shutdown()
                raise t.exception()

        task.add_done_callback(task_done_cb)

    def shutdown(self) -> None:
        if self._process:
            self._process.terminate()

    async def get_exchange_entities(
        self,
        on_response: Callable[..., Awaitable]
    ) -> MessageConsumerCollection:
        request = GetExchangeEntitiesRequest(
            self.exchange_info_dsn,
            self.exchange_info_get_entities_timeout
        )
        result = self._response_router.prepare_consumers_of_response(request)
        result.add_consumer(on_response)
        self._parent_connection.send(request)

        return result

    async def consume_price_depth(
        self,
        on_response: Callable[..., Awaitable],
        consume_keys: Optional[List[DepthConsumeKey]] = None
    ) -> MessageConsumerCollection:
        request = ConsumePriceDepthRequest(
            self.depth_scraping_queue_dsn,
            self.depth_scraping_queue_exchange,
            frozenset(consume_keys)
        )
        result = self._response_router.prepare_consumers_of_response(request)
        result.add_consumer(on_response)
        self._parent_connection.send(request)

        return result


class RealTransportProcess(Process):
    def __init__(
        self,
        *args,
        loop_debug: Optional[bool] = None,
        ready_event: Event,
        connection: Connection,
        poll_delay: float = 0.001,
        **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._loop_debug = loop_debug
        self._ready_event = ready_event
        self._connection = connection
        self._poll_delay = float(poll_delay)
        self._handlers: MutableMapping[Type[PipeRequest], Callable[..., Awaitable]] = {}

        self.add_handler(GetExchangeEntitiesRequest, self._get_exchange_entities)
        self.add_handler(ConsumePriceDepthRequest, self._consume_price_depth)

    @property
    def poll_delay(self):
        return self._poll_delay

    def add_handler(
        self,
        request_type: Type[PipeRequest],
        handler: Callable[..., Awaitable]
    ) -> None:
        if request_type in self._handlers:
            raise ValueError(f'Handler for {request_type} already registered')

        self._handlers[request_type] = handler

    def run(self) -> None:
        run_program_forever(self.main, loop_debug=self._loop_debug)

    async def main(self, program_env: AsyncProgramEnv) -> None:
        def exception_handler(_, context: Dict) -> None:
            if 'exception' in context:
                self._notify_owner_process(context['exception'])

        program_env.exception_handler_patch = exception_handler

        self._ready_event.set()

        while True:
            if not self._connection.poll():
                await asyncio.sleep(self.poll_delay)
                continue

            request = self._connection.recv()

            handler = self._find_handler_for_request(request)
            # @TODO The tasks should be collected for cancellation when this main task has been
            # cancelled. But when I've tried to done this I've got Segmentation fault from
            # unit-tests. May be it depends on Python version.
            asyncio.create_task(handler(request))

    async def _get_exchange_entities(self, request: GetExchangeEntitiesRequest) -> None:
        client = ExchangeInfoClient.factory(request.dsn, timeout_get_entities=request.timeout)
        entities = client.get_entities(generate_request_id())
        self._respond_to_owner_request(request, entities)

        client.destroy()

    async def _consume_price_depth(self, request: ConsumePriceDepthRequest) -> None:
        connection = RabbitConnection(request.dsn)
        channel = await connection.create_channel(100)
        cb = partial(self._price_depth_callback, request)
        consumer = RabbitConsumer(channel, request.exchange, cb)
        queue = await consumer.create_queue()

        if not request.consume_keys:
            routing_keys = ['#']
        else:
            routing_keys = [k.format_for_rabbitmq() for k in request.consume_keys]

        for routing_key in routing_keys:
            await queue.bind(consumer.exchange, routing_key)

    def _respond_to_owner_request(self, request: PipeRequest, content: Any) -> None:
        self._connection.send([request, content])

    def _notify_owner_process(self, original_exception: Exception) -> None:
        # This wrapping is required to don't fire unnecessary errors about serialization
        # of the exception. Otherwise a framework user will see undesired spam about
        # pickling RLock etc in logs.
        wrapped_exception = TransportFactoryException('An error in the transport process')
        # Unfortunately this line has no side-effects in Python 3.7 because this attribute
        # won't be packed by Connection.send() therefore cross-Connection.recv() will unpack
        # the data without the value for this attribute. So we don't have real wrapping here.
        wrapped_exception.__cause__ = original_exception
        self._connection.send([wrapped_exception])

    def _find_handler_for_request(self, request: PipeRequest) -> Callable[..., Awaitable]:
        request_type = type(request)
        if request_type not in self._handlers:
            raise ValueError(f'No handler for request type {request_type}')

        return self._handlers[request_type]

    def _price_depth_callback(
        self,
        request: ConsumePriceDepthRequest,
        message: aio_pika.IncomingMessage
    ) -> None:
        body = json.loads(message.body)

        for depth in body['depth'].values():
            for price_level in depth:
                rate = Decimal(price_level[0])
                amount = Decimal(price_level[1])

                if len(price_level) > 2 and price_level[2] is not None:
                    fee = Decimal(price_level[2])
                else:
                    fee = None

                price_level[:] = [rate, amount, fee]

        args = [
            body['exchange'],
            body['market'],
            body['symbol'],
            datetime.datetime.fromisoformat(body['now']),
            body['depth']['bids'],
            body['depth']['asks'],
        ]

        self._respond_to_owner_request(request, args)



from glob import glob
from os.path import basename, splitext

import setuptools

with open('README.md', 'r') as fh:
    long_description = fh.read()

setuptools.setup(
    name='galts-trade-api',
    version='0.1.0',
    license='Mozilla Public License 2.0',
    author='Sergey Nevmerzhitsky',
    author_email='sergey.nevmerzhitsky@gmail.com',
    description='A framework to use a trading infrastructure',
    long_description=long_description,
    long_description_content_type='text/markdown',
    keywords=['crypto', 'trading'],
    python_requires='~=3.7',
    install_requires=[
        'aio-pika~=5.0',
        'grpcio~=1.20',
        'protobuf~=3.11',  # Required by grpc_utils
        'structlog~=19.2.0',
        'ujson~=1.35',
    ],
    packages=setuptools.find_packages('src'),
    # https://packaging.python.org/guides/distributing-packages-using-setuptools/#py-modules
    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],
    package_dir={'': 'src'},
    include_package_data=True,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Financial and Insurance Industry',
        'License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
    ],
)



import datetime
from typing import Optional


class Asset:
    def __init__(
        self,
        id: int,
        tag: str,
        name: str,
        precision: int,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime]
    ):
        self._id: int = int(id)
        self._tag: str = str(tag).strip()
        self._name: str = str(name).strip()
        self._precision: int = int(precision)
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time

    @property
    def id(self):
        return self._id

    @property
    def tag(self):
        return self._tag

    @property
    def name(self):
        return self._name

    @property
    def precision(self):
        return self._precision

    @property
    def create_time(self):
        return self._create_time

    @property
    def delete_time(self):
        return self._delete_time


class Symbol:
    @classmethod
    def form_tag(cls, base_asset_tag: str, quote_asset_tag: str) -> str:
        return f'{base_asset_tag}{quote_asset_tag}'

    def __init__(
        self,
        id: int,
        base_asset_id: int,
        quote_asset_id: int,
        create_time: datetime.datetime,
        delete_time: Optional[datetime.datetime]
    ):
        self._id: int = int(id)
        self._base_asset_id: int = int(base_asset_id)
        self._quote_asset_id: int = int(quote_asset_id)
        self._create_time: datetime.datetime = create_time
        self._delete_time: Optional[datetime.datetime] = delete_time

    @property
    def id(self):
        return self._id

    @property
    def base_asset_id(self):
        return self._base_asset_id

    @property
    def quote_asset_id(self):
        return self._quote_asset_id

    @property
    def create_time(self):
        return self._create_time

    @property
    def delete_time(self):
        return self._delete_time



import logging
import sys
from typing import Dict, Mapping, Optional, TextIO

import structlog
import ujson
from structlog.processors import _figure_out_exc_info

__all__ = ['get_logger', 'configure_json_output', 'UltraJSONRenderer']


def _shy_format_exc_info(_, __, event_dict: Dict) -> Mapping:
    """
    A version of `structlog.processors.format_exc_info` which doesn't dump a tracback.
    Check http://www.structlog.org/en/19.1.0/processors.html for info about processors.
    """
    exc_info = event_dict.pop('exc_info', None)
    if exc_info:
        exc_tuple = _figure_out_exc_info(exc_info)
        event_dict['exceptionClass'] = exc_tuple[0].__name__
        event_dict['exceptionMessage'] = str(exc_tuple[1])
    return event_dict


class UltraJSONRenderer(structlog.processors.JSONRenderer):
    def __init__(self, **dumps_kw):
        self._dumps_kw = dumps_kw
        self._dumps = ujson.dumps


def configure_json_output(enable_unicode_decoder: bool = False, **json_dumps_kwargs) -> None:
    """
    Args:
         json_dumps_kwargs: Optional arguments for JSON serializer. Useful value in development:
            {'indent': 2}
    """
    if structlog.is_configured():
        return

    json_dumps_kwargs.setdefault('escape_forward_slashes', False)

    processors = [
        structlog.stdlib.filter_by_level,
        structlog.stdlib.add_logger_name,
        structlog.stdlib.add_log_level,
        structlog.stdlib.PositionalArgumentsFormatter(),
        structlog.processors.TimeStamper('iso', True),
        structlog.processors.StackInfoRenderer(),
        _shy_format_exc_info,
    ]

    if enable_unicode_decoder:
        processors.append(structlog.processors.UnicodeDecoder())

    processors.append(UltraJSONRenderer(**json_dumps_kwargs))

    structlog.configure_once(
        processors=processors,
        logger_factory=structlog.stdlib.LoggerFactory(),
        wrapper_class=structlog.stdlib.BoundLogger,
        # Like http://www.structlog.org/en/19.1.0/performance.html advice
        cache_logger_on_first_use=True
    )


def get_logger(
    name: Optional[str] = None,
    level: Optional[int] = logging.INFO,
    stream: Optional[TextIO] = sys.stdout,
    *args,
    **kwargs
) -> logging.Logger:
    # To correct work of stdlib we should initialize the standard logger with the name before
    # interaction with structlog
    std_logger = logging.getLogger(name)
    if level is not None:
        std_logger.setLevel(level)
    if stream is not None:
        std_logger.addHandler(logging.StreamHandler(stream))

    return structlog.get_logger(name, *args, **kwargs)
